% ----------------------------------------------------------------------------
% 
% Copyright (c) 2016 Gilles Simon (gsimon[at]loria.fr)
% 
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
% 
% This is an implementation of the method described in: 
% 
% Gilles Simon, Antoine Fond, Marie-Odile Berger 
% A Simple and Effective Method to Detect Orthogonal Vanishing Points in Uncalibrated Images of Man-Made Environments. 
% Eurographics 2016, May 2016, Lisbon, Portugal.
% 
% ----------------------------------------------------------------------------

function [horizon_line, horizon_vp, zenith, K, vp_association, hrank, times] = fast_vp(im, lines0, params)
%% parameters setup & computation

n_candidate = 32;
roll_max = pi/32;
angular_accuracy = 1;
infty = 64;%32;
one_deg = pi/180;
sig_consistency = 0.5;
thres_consistency = exp(-1/2);
sensitivity = min(3,params.sensitivity);
peak_thres = 2^(5-sensitivity);

horizon_line = [];
horizon_vp = [];
zenith = [];
vp_association = [];
hrank = -1;

width = size(im, 2);
height = size(im, 1);
nt = 2^(5+min(params.accuracy, log2(width)-5));
step_h = width/nt;
step_v = height/nt;
medfilt_n = round(width/step_h);
margin = round(medfilt_n/4);
K = [];
u0 = width/2;
v0 = height/2;
Dt_h = atan(step_h/width);
x_axis = [];
y_axis = [];
imax = ceil(pi/2/Dt_h);
for i=-imax:imax
    x_axis(i+imax+1) = u0 + width*tan(i*Dt_h);
end
Dt_v = atan(step_v/height);
imaxv = ceil(pi/2/Dt_v);
for i=-imaxv:imaxv
    y_axis(i+imaxv+1) = height*tan(i*Dt_v);
end

%% estimation of the vertical vanishing point (Zenith) & image rotation

t = tic;
angle0=[];
for i = 1:size(lines0,1)
    centroid0(i,1) = (lines0(i,1)+lines0(i,3))/2;
    centroid0(i,2) = (lines0(i,2)+lines0(i,4))/2;
    angle0(i) = line_angle(lines0(i,:));
    L(i) = sqrt((lines0(i,1)-lines0(i,3))^2+(lines0(i,2)-lines0(i,4))^2);
    lines0I(i,1) = (lines0(i,3)-lines0(i,1))/L(i);
    lines0I(i,2) = (lines0(i,4)-lines0(i,2))/L(i);
    lines0J(i,1) = -lines0I(i,2);
    lines0J(i,2) = lines0I(i,1);    
end
clear vvp_hist;
h = 0;
for theta = (pi/2-roll_max):(angular_accuracy*one_deg):(pi/2+roll_max)
    clear vvp_cpt;
    id = 0;
    for ad = y_axis
        if ad < -height/2 || ad > height/2
            id = id + 1;
            u = u0 + ad*cos(theta);
            v = v0 + ad*sin(theta);
            vvp_cpt(id,1) = 0;
            for j = 1:size(lines0,1)
                vp_line(1, 1) = u;
                vp_line(1, 2) = v;
                vp_line(1, 3) = centroid0(j, 1);
                vp_line(1, 4) = centroid0(j, 2);
                vp_angle = line_angle(vp_line);
                if (abs(vp_angle-angle0(j)) < angular_accuracy*one_deg/2)
                    vvp_cpt(id,1) = vvp_cpt(id,1) + 1;
                end
            end
            vvp_cpt(id,2) = u;
            vvp_cpt(id,3) = v;
        end
    end
    [vvp_cpt_max, id_vvp_cpt_max] = max(vvp_cpt(:,1));
    h = h+1;
    vvp_hist(h,1) = vvp_cpt(id_vvp_cpt_max,2);
    vvp_hist(h,2) = vvp_cpt(id_vvp_cpt_max,3);
    vvp_hist(h,3) = theta;
    vvp_hist(h,4) = vvp_cpt(id_vvp_cpt_max,1);
end
[vvp_hist_max, id_vvp_hist_max] = max(vvp_hist(:,4));
zenith0.u = vvp_hist(id_vvp_hist_max,1);
zenith0.v = vvp_hist(id_vvp_hist_max,2);
zenith0.count = vvp_hist(id_vvp_hist_max,4);
rot_angle = (vvp_hist(id_vvp_hist_max,3) - pi/2);
if rot_angle < - pi/4
    rot_angle = rot_angle + pi/2;
end
lines = [];
angle = [];
centroid = [];
for i = 1:size(lines0,1)
    lines(i,:) = line_rotate(lines0(i,:), -rot_angle, u0, v0);
%     L(i) = sqrt((lines(i,1)-lines(i,3))^2+(lines(i,2)-lines(i,4))^2);
    linesI(i,1) = (lines(i,3)-lines(i,1))/L(i);
    linesI(i,2) = (lines(i,4)-lines(i,2))/L(i);
    linesJ(i,1) = -linesI(i,2);
    linesJ(i,2) = linesI(i,1);    
    centroid(i,1) = (lines(i,1)+lines(i,3))/2;
    centroid(i,2) = (lines(i,2)+lines(i,4))/2;
    angle(i) =  angle0(i) + rot_angle;
end

t = toc(t); disp(['[zenith] ',num2str(t),' seconds elapsed.']);
times(1,1) = t;

%% computation of the horizon candidates and vp maxima along these lines

t = tic;

zenith = point_rotate([zenith0.u,zenith0.v], -rot_angle, u0, v0);
y = [];
nedges = 0;
for i = 1:size(lines,1)
    if abs(angle(i)) < angular_accuracy*one_deg
        nedges = nedges+1;
        y(nedges) = centroid(i,2);
    end
end
if nedges == 0
    nedges = 1;
    y(nedges) = v0;
end
nbins = round(height/4);
[hl_hist, hl_edges] = histcounts(y,nbins);
[hmax, hmax_pos] = sort(hl_hist, 'descend');

t = toc(t); disp(['[histogram] ',num2str(t),' seconds elapsed.']);
times(1,2) = t;

t = tic;

counter = cell(1,n_candidate);
maxima = [];
for i = 1:min(n_candidate,max(find(hmax~=0)))
    v_candidate = (hl_edges(hmax_pos(i)) + hl_edges(hmax_pos(i)+1))/2;
    id = 0;
    for u = x_axis
        id = id + 1;
        counter{i}(id,1) = 0;
        for j = 1:size(lines,1)
            vp_vec(1,1) = u-lines(j,1);
            vp_vec(1,2) = v_candidate-lines(j,2);
            u_v = vp_vec(1,1)*linesI(j,1)+vp_vec(1,2)*linesI(j,2);
            v_v = vp_vec(1,1)*linesJ(j,1)+vp_vec(1,2)*linesJ(j,2);
            consistency = exp(-v_v^2*L(j)^2/2/sig_consistency^2/(u_v^2+(u_v-L(j))^2));
            if (abs(abs(angle(j))-pi/2) > pi/4 && consistency > thres_consistency)
                counter{i}(id,1) = counter{i}(id,1) + 1;
            end
        end
        counter{i}(id,2) = u;
    end
    ct = counter{i}(:,1);
    half_medfilt_n = round(medfilt_n/2);
    ct_ext(1:half_medfilt_n,1) = ct((end-half_medfilt_n+1):end);
    ct_ext((half_medfilt_n+1):(half_medfilt_n+length(ct)),1) = ct;
    ct_ext((half_medfilt_n+length(ct)+1):(half_medfilt_n+length(ct)+half_medfilt_n),1) = ct(1:half_medfilt_n,1);
    medf = medfilt1(ct_ext, medfilt_n);
    medf = medf((half_medfilt_n+1):(half_medfilt_n+length(ct)));
    diff = ct-medf;
    diff = diff.*double(diff > 0);
    diff = diff.*double(diff > peak_thres);
    [maxima(i,3), id_ct_max] = max(ct);
    maxima(i,1) = v_candidate;
    maxima(i,2) = 1;
    maxima(i,4) = counter{i}(id_ct_max,2);
    idmin = id_ct_max-half_medfilt_n;
    idmax = id_ct_max+half_medfilt_n;
    if idmin < 1
        for k = (length(diff)+idmin):length(diff)
            diff(k) = 0;
        end
    end
    if idmax > length(diff)
        for k = 1:(idmax-length(diff))
            diff(k) = 0;
        end
    end
    for k = max(1,idmin):min(length(diff),idmax)
        diff(k) = 0;
    end
    while max(diff) > 0
        [lmax, lmax_id] = max(diff);
        u_lmax = counter{i}(lmax_id,2);
        maxima(i,2) = maxima(i,2) +1;
        if maxima(i,2) == 2
            maxima(i,3) = maxima(i,3) + ct(lmax_id);
        end
        maxima(i,3+maxima(i,2)) = u_lmax;
        idmin = lmax_id-half_medfilt_n;
        idmax = lmax_id+half_medfilt_n;
        if idmin < 1
            for k = (length(diff)+idmin):length(diff)
                diff(k) = 0;
            end
        end
        if idmax > length(diff)
            for k = 1:(idmax-length(diff))
                diff(k) = 0;
            end
        end
        for k=max(1,idmin):min(length(diff),idmax)
            diff(k) = 0;
        end
     end
end
[count_max, hrank] = max(maxima(:,3));
best_horizon_vp = maxima(hrank,:);
horizon_vp_count = best_horizon_vp(2);
horizon_vp = [];
for v = 1:horizon_vp_count
    horizon_vp(v,1) = best_horizon_vp(3+v);
    horizon_vp(v,2) = best_horizon_vp(1);
end

t = toc(t); disp(['[horizontal vp] ',num2str(t),' seconds elapsed.']);
times(1,3) = t;

%% Manhattan direction estimate

t = tic;

zenith_candidates_count = 0;
zenith_candidates = [];
fc = [];
f = [];
zenith_id = 0;
if horizon_vp_count > 1 && abs(zenith(2) - v0) < infty*height
    for v = 1:horizon_vp_count
        for v2 = (v+1):horizon_vp_count
            x1 = horizon_vp(v,1) - u0;
            y1 = horizon_vp(v,2) - v0;
            x2 = horizon_vp(v2,1) - u0;
            y2 = horizon_vp(v2,2) - v0;
            fc(v,v2) = sqrt(-(x1*x2+y1*y2));
            ff = fc(v,v2);
            if ff*ff > 0 && ff > 0.28*width && ff < 3.8*width
                zenith_candidates_count = zenith_candidates_count + 1;
                vv1 = [x1/ff;y1/ff;1];
                vv2 = [x2/ff;y2/ff;1];
                vv3 = cross(vv1,vv2);
                zenith_candidates(zenith_candidates_count, 1) = ff*vv3(1)/vv3(3)+u0;
                zenith_candidates(zenith_candidates_count, 2) = ff*vv3(2)/vv3(3)+v0;
                zenith_candidates(zenith_candidates_count, 3) = ff;
                zenith_candidates(zenith_candidates_count, 5) = v;
                zenith_candidates(zenith_candidates_count, 6) = v2;
                at_vz1 = atan(zenith_candidates(zenith_candidates_count, 2)/double(height))/Dt_h;
                at_vz2 = atan(zenith(2)/double(height))/Dt_h;
                zenith_candidates(zenith_candidates_count, 4) = abs(at_vz1-at_vz2);
            end
        end
    end
end
if zenith_candidates_count > 0
    [vvpc_min, id_vvpc_min] = min(zenith_candidates(:,4));
    if vvpc_min < 4
        zenith_id = id_vvpc_min;
        id1 = zenith_candidates(zenith_id, 5);
        id2 = zenith_candidates(zenith_id, 6);
        x1 = horizon_vp(id1,1);
        x2 = horizon_vp(id2,1);
        y = zenith_candidates(zenith_id, 2);
        if (y > v0 && x1 > x2) || (y < v0 && x1 < x2)
            zenith_candidates(zenith_id, 5) = id2;
            zenith_candidates(zenith_id, 6) = id1;
        end
        ff = zenith_candidates(zenith_id,3);
        K = [[ff 0 u0];[0 ff v0];[0 0 1]];
    end
end
if zenith_id == 0
    zenith_id = 1;
    zenith_candidates = point_rotate([zenith0.u,zenith0.v], -rot_angle, u0, v0);
    zenith_candidates(1,3:6) = 0;
    if horizon_vp_count > 0 && abs(horizon_vp(1,1)-u0) < infty*width && abs(zenith_candidates(1,2) - v0) < infty*height
        y1 = horizon_vp(1,2) - v0;
        y2 = zenith_candidates(1,2) - v0;
        ff = sqrt(-y1*y2);
        if ff*ff > 0 && ff > 0.28*width && ff < 3.8*width
            K = [[ff 0 u0];[0 ff v0];[0 0 1]];
        end
    end
end
zenith(1,1) = zenith_candidates(zenith_id, 1);
zenith(1,2) = zenith_candidates(zenith_id, 2);
zenith(1,3) = zenith_candidates(zenith_id, 5);
zenith(1,4) = zenith_candidates(zenith_id, 6);
zenith(1,1:2) = point_rotate(zenith(1,1:2), rot_angle, u0, v0);
for v = 1:horizon_vp_count
    horizon_vp(v,:) = point_rotate(horizon_vp(v,:), rot_angle, u0, v0);
end
hlp1(1,1) = 1;
hlp1(1,2) = best_horizon_vp(1);
hlp2(1,1) = width;
hlp2(1,2) = best_horizon_vp(1);
hlp1_r = point_rotate(hlp1, rot_angle, u0, v0);
hlp2_r = point_rotate(hlp2, rot_angle, u0, v0);
horizon_line = line_hmg_from_two_points(hlp1_r,hlp2_r);

t = toc(t); disp(['[Manhattan directions] ',num2str(t),' seconds elapsed.']);
times(1,4) = t;

if params.plot
    fig = figure();
    hold on;
    s = length(counter{hrank}(:,1));
    hs = floor(s/2);
    plot([-hs:s-hs-1]',counter{hrank}(:,1));
    for ab=-imax*width:width:imax*width
        id = round(atan(ab/width - u0/width)/Dt_h +imax+1);
        plot([id-hs-1;id-hs-1],[0;10],'m');
    end
    for v = 1:horizon_vp_count
        uu = horizon_vp(v,1);
        id = round(atan(uu/width - u0/width)/Dt_h +imax+1);
        if (zenith(1,3) == 0 && v == 1) || (v == zenith(1,3))
            plot([id-hs-1; id-hs-1],[0;counter{hrank}(id)], '-r', 'LineWidth', 2);
        else
            if (v == zenith(1,4))
                plot([id-hs-1; id-hs-1],[0;counter{hrank}(id)], '-g', 'LineWidth', 2);
            else
                ax = gca;
                ax.ColorOrderIndex = v;
                plot([id-hs-1; id-hs-1],[0;counter{hrank}(id)], '-', 'LineWidth', 2);
            end
        end
    end
    drawnow();
    fprintf('=> %s\n', params.s_fname);
    hgexport(fig, params.s_fname);
end

%% get lines - vps associations

vp_association = -1*ones(size(lines0,1),1);
for i = 1:size(lines0,1)
    bundle_line = [zenith(1,1) zenith(1,2) centroid0(i,1) centroid0(i,2)];
    d_angle = line_angle(bundle_line);
    if (abs(d_angle-angle0(i)) < angular_accuracy*one_deg)
        vp_association(i) = 0;
    else
        for v = 1:horizon_vp_count
            vp_vec(1,1) = horizon_vp(v,1)-lines0(i,1);
            vp_vec(1,2) = horizon_vp(v,2)-lines0(i,2);
            u_v = vp_vec(1,1)*lines0I(i,1)+vp_vec(1,2)*lines0I(i,2);
            v_v = vp_vec(1,1)*lines0J(i,1)+vp_vec(1,2)*lines0J(i,2);
            consistency = exp(-v_v^2*L(i)^2/2/sig_consistency^2/(u_v^2+(u_v-L(i))^2));
            if (consistency > thres_consistency)
                vp_association(i) = v;
            end
            
        end
    end
end
end
