% ----------------------------------------------------------------------------
% 
% Copyright (c) 2016 Gilles Simon (gsimon[at]loria.fr)
% 
% This program is free software; you can redistribute it and/or modify
% it under the terms of the GNU General Public License as published by
% the Free Software Foundation; either version 3 of the License, or
% (at your option) any later version.
% 
% This program is distributed in the hope that it will be useful,
% but WITHOUT ANY WARRANTY; without even the implied warranty of
% MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
% GNU Affero General Public License for more details.
% 
% This is an implementation of the method described in: 
% 
% Gilles Simon, Antoine Fond, Marie-Odile Berger 
% A Simple and Effective Method to Detect Orthogonal Vanishing Points in Uncalibrated Images of Man-Made Environments. 
% Eurographics 2016, May 2016, Lisbon, Portugal.
% 
% ----------------------------------------------------------------------------

clear all; clc;
addpath('./lsd-1.5/');

params.accuracy = 2; % int between 0 (fast but not accurate) and log2(width)-5 (accurate but slow)
params.sensitivity = 2; % int bewtween 0 (only obvious vps are detected) and 3 (very sensitive, may lead to spurious vps)
params.plot = true

multi_frame = true;
reduce = true;
reduced_size = 720;
do_orthorectify = true;
PLOT = true;

if multi_frame
    img_path = uigetdir();
    files = dir(strcat(img_path, '/*.jpg'));
else
    [files(1).name, img_path] = uigetfile('*.*');
end
if params.plot
    params.fig = figure();
end
dstr = datestr(datetime);
pathres = strcat(img_path, '/res_', dstr);
cmd = strcat('mkdir ''',img_path, '/res_', dstr,'''');
system(cmd); pause(4);

time_frame = zeros(size(files,1),4);

for im_num = 1:length(files)
    close all;
     
    fname = fullfile(img_path, files(im_num).name)
    [pathstr, name, ext] = fileparts(fname);
    
    if params.plot
        params.s_fname = strcat(pathres,'/',name,'_lcount.eps');
    end
    
    t0 = tic();
    
    % get line segments from lsd
    
    t = tic();

    im=imread(fname);
    if reduce && size(im,2) > reduced_size
        ratio = reduced_size/size(im,2);
        tform = affine2d([ratio 0 0;0 ratio 0;0 0 1]);
        im = imwarp(im,tform,'cubic');
    end
    gim=im;
    gim = rgb2gray(im);
    lines = lsd(double(gim));
    
    t = toc(t); disp(['[lsd] ',num2str(t),' seconds elapsed.']);
    time_frame(im_num, 1) = t;
    
    % vp extraction
    
    [horizon_line, horizon_vp, zenith, K, vp_association, hrank, time_frame(im_num,2:5)] = fast_vp(im, lines, params);

    time_frame(im_num,6) = toc(t0); disp(['[TOTAL TIME] ',num2str(time_frame(im_num,6)),' seconds elapsed.']);
    s_fname = strcat(pathres,'/time_frame.mat');
    save(s_fname, 'time_frame');
    
    s_fname = strcat(pathres,'/',name,'_vp.mat');
    fprintf('=> %s\n', s_fname);
    save(s_fname, 'fname', 'horizon_line', 'horizon_vp', 'zenith', 'K', 'lines', 'vp_association', 'hrank');

    if PLOT && length(horizon_line) == 3
        fig = vp_plot(im, horizon_line, (length(K) > 0), horizon_vp, zenith, lines, vp_association);
        s_fname = strcat(pathres,'/',name,'_result');
        fprintf('=> %s\n', s_fname);
        hgexport(fig, s_fname);
        savefig(fig, s_fname);     
    end

    % image rectification from vps
    
    if do_orthorectify
        [imR,H] = orthorectify_vp(im, horizon_vp, zenith, lines, vp_association, 8, K, horizon_line, params.plot);
        for i = 1:length(imR)
            s_fname = strcat(pathres,'/',name,'_rec',num2str(i),'.jpg');
            fprintf('=> %s\n', s_fname);
            imwrite(imR{i}, s_fname);
            if PLOT && ~params.plot
                figure; imshow(imR{i}); drawnow();
            end
        end
    end
end


